@@ -18,7 +18,7 @@
 
 #include "client_priv.h"
 #include <sslopt-vars.h>
-#include <../scripts/mysql_fix_privilege_tables_sql.c>
+#include "../scripts/mysql_fix_privilege_tables_sql.c"
 
 #include <welcome_copyright_notice.h> /* ORACLE_WELCOME_COPYRIGHT_NOTICE */
 
@@ -153,7 +153,7 @@ static struct my_option my_long_options[]=
    &opt_systables_only, &opt_systables_only, 0,
    GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
 #define USER_OPT (array_elements(my_long_options) - 6)
-  {"user", 'u', "User for login.", &opt_user,
+  {"user", 'u', "User for login if not current user.", &opt_user,
    &opt_user, 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
   {"verbose", 'v', "Display more output about the process; Using it twice will print connection argument; Using it 3 times will print out all CHECK, RENAME and ALTER TABLE during the check phase.",
    &opt_not_used, &opt_not_used, 0, GET_BOOL, NO_ARG, 1, 0, 0, 0, 0, 0},
@@ -558,8 +558,7 @@ static int run_query(const char *query, DYNAMIC_STRING *ds_res,
   DBUG_PRINT("enter", ("query: %s", query));
   if ((fd= create_temp_file(query_file_path, 
                             opt_tmpdir[0] ? opt_tmpdir : NULL,
-                            "sql", O_CREAT | O_SHARE | O_RDWR,
-                            MYF(MY_WME))) < 0)
+                            "sql", O_SHARE, MYF(MY_WME))) < 0)
     die("Failed to create temporary file for defaults");
 
   /*
@@ -673,7 +672,7 @@ static int get_upgrade_info_file_name(char* name)
 
 */
 
-static int upgrade_already_done(void)
+static int upgrade_already_done(myf flags)
 {
   FILE *in;
   char upgrade_info_file[FN_REFLEN]= {0};
@@ -681,17 +680,23 @@ static int upgrade_already_done(void)
   if (get_upgrade_info_file_name(upgrade_info_file))
     return 0; /* Could not get filename => not sure */
 
-  if (!(in= my_fopen(upgrade_info_file, O_RDONLY, MYF(0))))
+  if (!(in= my_fopen(upgrade_info_file, O_RDONLY, flags)))
     return 0; /* Could not open file => not sure */
 
   bzero(upgrade_from_version, sizeof(upgrade_from_version));
   if (!fgets(upgrade_from_version, sizeof(upgrade_from_version), in))
   {
-    /* Ignore, will be detected by strncmp() below */
+    /* Preserve errno for caller */
+    int save_errno= errno;
+    (void) my_fclose(in, flags);
+    errno= save_errno;
+    return 0;
   }
 
-  my_fclose(in, MYF(0));
+  if (my_fclose(in, flags))
+    return 0;
 
+  errno= 0;
   return (strncmp(upgrade_from_version, MYSQL_SERVER_VERSION,
                   sizeof(MYSQL_SERVER_VERSION)-1)==0);
 }
@@ -720,23 +725,24 @@ static void create_mysql_upgrade_info_file(void)
   {
     fprintf(stderr,
             "Could not create the upgrade info file '%s' in "
-            "the MySQL Servers datadir, errno: %d\n",
+            "the MariaDB Servers datadir, errno: %d\n",
             upgrade_info_file, errno);
     return;
   }
 
   /* Write new version to file */
-  fputs(MYSQL_SERVER_VERSION, out);
-  my_fclose(out, MYF(0));
+  my_fwrite(out, (uchar*) MYSQL_SERVER_VERSION,
+            sizeof(MYSQL_SERVER_VERSION), MY_WME);
+  my_fclose(out, MYF(MY_WME));
 
   /*
     Check if the upgrad_info_file was properly created/updated
     It's not a fatal error -> just print a message if it fails
   */
-  if (!upgrade_already_done())
+  if (!upgrade_already_done(MY_WME))
     fprintf(stderr,
-            "Could not write to the upgrade info file '%s' in "
-            "the MySQL Servers datadir, errno: %d\n",
+            "Upgrade file '%s' was not properly created. "
+            "Got error errno while checking file content: %d\n",
             upgrade_info_file, errno);
   return;
 }
@@ -958,8 +964,9 @@ static int install_used_engines(void)
 {
   char buf[512];
   DYNAMIC_STRING ds_result;
-  const char *query = "SELECT DISTINCT LOWER(engine) FROM information_schema.tables"
-                      " WHERE table_comment LIKE 'Unknown storage engine%'";
+  const char *query = "SELECT DISTINCT LOWER(engine) AS c1 FROM information_schema.tables"
+                      " WHERE table_comment LIKE 'Unknown storage engine%'"
+                      " ORDER BY c1";
 
   if (opt_systables_only || !from_before_10_1())
   {
@@ -1194,7 +1201,7 @@ int main(int argc, char **argv)
   load_defaults_or_exit("my", load_default_groups, &argc, &argv);
   defaults_argv= argv; /* Must be freed by 'free_defaults' */
 
-#if __WIN__
+#if defined(__WIN__)
   if (GetModuleFileName(NULL, self_name, FN_REFLEN) == 0)
 #endif
   {
@@ -1226,7 +1233,7 @@ int main(int argc, char **argv)
   cnf_file_path= strmov(defaults_file, "--defaults-file=");
   {
     int fd= create_temp_file(cnf_file_path, opt_tmpdir[0] ? opt_tmpdir : NULL,
-                             "mysql_upgrade-", O_CREAT | O_WRONLY, MYF(MY_FAE));
+                             "mysql_upgrade-", 0, MYF(MY_FAE));
     if (fd < 0)
       die(NULL);
     my_write(fd, USTRING_WITH_LEN( "[client]\n"), MYF(MY_FAE));
@@ -1247,7 +1254,7 @@ int main(int argc, char **argv)
     Read the mysql_upgrade_info file to check if mysql_upgrade
     already has been run for this installation of MySQL
   */
-  if (!opt_force && upgrade_already_done())
+  if (!opt_force && upgrade_already_done(0))
   {
     printf("This installation of MySQL is already upgraded to %s, "
            "use --force if you still need to run mysql_upgrade\n",
