@@ -116,17 +116,23 @@ my_hash_init2(HASH *hash, uint growth_size, CHARSET_INFO *charset,
 static inline void my_hash_free_elements(HASH *hash)
 {
   uint records= hash->records;
+  if (records == 0)
+    return;
+
   /*
     Set records to 0 early to guard against anyone looking at the structure
     during the free process
   */
   hash->records= 0;
+
   if (hash->free)
   {
     HASH_LINK *data=dynamic_element(&hash->array,0,HASH_LINK*);
     HASH_LINK *end= data + records;
-    while (data < end)
+    do
+    {
       (*hash->free)((data++)->data);
+    } while (data < end);
   }
 }
 
@@ -380,7 +386,7 @@ static int hashcmp(const HASH *hash, HASH_LINK *pos, const uchar *key,
 my_bool my_hash_insert(HASH *info, const uchar *record)
 {
   int flag;
-  uint idx, halfbuff, first_index;
+  size_t idx, halfbuff, first_index;
   size_t length;
   my_hash_value_type current_hash_nr, UNINIT_VAR(rec_hash_nr),
     UNINIT_VAR(rec2_hash_nr);
