@@ -31,7 +31,7 @@
 #pragma implementation				// gcc: Class implementation
 #endif
 
-#include <my_global.h>
+#include "mariadb.h"
 #include "sql_priv.h"
 #include "unireg.h"
 #include "sql_base.h"                           // close_mysql_tables
@@ -52,16 +52,14 @@ static bool initialized = 0;
 static MEM_ROOT mem;
 static HASH udf_hash;
 static mysql_rwlock_t THR_LOCK_udf;
+static LEX_CSTRING MYSQL_FUNC_NAME= {STRING_WITH_LEN("func") };
 
-
-static udf_func *add_udf(LEX_STRING *name, Item_result ret,
-                         char *dl, Item_udftype typ);
+static udf_func *add_udf(LEX_CSTRING *name, Item_result ret,
+                         const char *dl, Item_udftype typ);
 static void del_udf(udf_func *udf);
 static void *find_udf_dl(const char *dl);
-static bool find_udf_everywhere(THD* thd, const char *name, uint len,
-                                TABLE *table);
 
-static char *init_syms(udf_func *tmp, char *nm)
+static const char *init_syms(udf_func *tmp, char *nm)
 {
   char *end;
 
@@ -144,7 +142,6 @@ void udf_init()
   TABLE *table;
   int error;
   DBUG_ENTER("ufd_init");
-  char db[]= "mysql"; /* A subject to casednstr, can't be constant */
 
   if (initialized || opt_noacl)
     DBUG_VOID_RETURN;
@@ -155,7 +152,7 @@ void udf_init()
 
   mysql_rwlock_init(key_rwlock_THR_LOCK_udf, &THR_LOCK_udf);
 
-  init_sql_alloc(&mem, UDF_ALLOC_BLOCK_SIZE, 0, MYF(0));
+  init_sql_alloc(&mem, "udf", UDF_ALLOC_BLOCK_SIZE, 0, MYF(0));
   THD *new_thd = new THD(0);
   if (!new_thd ||
       my_hash_init(&udf_hash,system_charset_info,32,0,0,get_hash_key, NULL, 0))
@@ -169,9 +166,9 @@ void udf_init()
   initialized = 1;
   new_thd->thread_stack= (char*) &new_thd;
   new_thd->store_globals();
-  new_thd->set_db(db, sizeof(db)-1);
+  new_thd->set_db(&MYSQL_SCHEMA_NAME);
 
-  tables.init_one_table(db, sizeof(db)-1, "func", 4, "func", TL_READ);
+  tables.init_one_table(&new_thd->db, &MYSQL_FUNC_NAME, 0, TL_READ);
 
   if (open_and_lock_tables(new_thd, &tables, FALSE, MYSQL_LOCK_IGNORE_TIMEOUT))
   {
@@ -191,10 +188,10 @@ void udf_init()
   }
 
   table->use_all_columns();
-  while (!(error= read_record_info.read_record(&read_record_info)))
+  while (!(error= read_record_info.read_record()))
   {
     DBUG_PRINT("info",("init udf record"));
-    LEX_STRING name;
+    LEX_CSTRING name;
     name.str=get_field(&mem, table->field[0]);
     name.length = (uint) strlen(name.str);
     char *dl_name= get_field(&mem, table->field[2]);
@@ -244,7 +241,8 @@ void udf_init()
     }
     tmp->dlhandle = dl;
     {
-      char buf[SAFE_NAME_LEN+16], *missing;
+      char buf[SAFE_NAME_LEN+16];
+      const char *missing;
       if ((missing= init_syms(tmp, buf)))
       {
         sql_print_error(ER_THD(new_thd, ER_CANT_FIND_DL_ENTRY), missing);
@@ -254,12 +252,10 @@ void udf_init()
       }
     }
   }
-  if (error > 0)
+  if (unlikely(error > 0))
     sql_print_error("Got unknown error: %d", my_errno);
   end_read_record(&read_record_info);
-
-  // Force close to free memory
-  table->mark_table_for_reopen();
+  table->m_needs_reopen= TRUE;                  // Force close to free memory
 
 end:
   close_mysql_tables(new_thd);
@@ -315,9 +311,9 @@ static void del_udf(udf_func *udf)
       The functions will be automaticly removed when the least threads
       doesn't use it anymore
     */
-    char *name= udf->name.str;
-    uint name_length=udf->name.length;
-    udf->name.str=(char*) "*";
+    const char *name= udf->name.str;
+    size_t name_length=udf->name.length;
+    udf->name.str= "*";
     udf->name.length=1;
     my_hash_update(&udf_hash,(uchar*) udf,(uchar*) name,name_length);
   }
@@ -351,10 +347,11 @@ void free_udf(udf_func *udf)
 
 /* This is only called if using_udf_functions != 0 */
 
-udf_func *find_udf(const char *name,uint length,bool mark_used)
+udf_func *find_udf(const char *name,size_t length,bool mark_used)
 {
   udf_func *udf=0;
   DBUG_ENTER("find_udf");
+  DBUG_ASSERT(strlen(name) == length);
 
   if (!initialized)
     DBUG_RETURN(NULL);
@@ -366,8 +363,7 @@ udf_func *find_udf(const char *name,uint length,bool mark_used)
   else
     mysql_rwlock_rdlock(&THR_LOCK_udf);  /* Called during parsing */
 
-  if ((udf=(udf_func*) my_hash_search(&udf_hash,(uchar*) name,
-                                      length ? length : (uint) strlen(name))))
+  if ((udf=(udf_func*) my_hash_search(&udf_hash,(uchar*) name, length)))
   {
     if (!udf->dlhandle)
       udf=0;					// Could not be opened
@@ -399,7 +395,7 @@ static void *find_udf_dl(const char *dl)
 
 /* Assume that name && dl is already allocated */
 
-static udf_func *add_udf(LEX_STRING *name, Item_result ret, char *dl,
+static udf_func *add_udf(LEX_CSTRING *name, Item_result ret, const char *dl,
 			 Item_udftype type)
 {
   if (!name || !dl || !(uint) type || (uint) type > (uint) UDFTYPE_AGGREGATE)
@@ -419,45 +415,6 @@ static udf_func *add_udf(LEX_STRING *name, Item_result ret, char *dl,
   return tmp;
 }
 
-/**
-  Find record with the udf in the udf func table
-
-  @param exact_name_str  udf name
-  @param exact_name_len  udf name length
-  @param table           table of mysql.func
-
-  @retval TRUE  found
-  @retral FALSE not found
-*/
-
-static bool find_udf_in_table(const char *exact_name_str, uint exact_name_len,
-                              TABLE *table)
-{
-  table->use_all_columns();
-  table->field[0]->store(exact_name_str, exact_name_len, &my_charset_bin);
-  return (!table->file->ha_index_read_idx_map(table->record[0], 0,
-                                              (uchar*) table->field[0]->ptr,
-                                              HA_WHOLE_KEY,
-                                              HA_READ_KEY_EXACT));
-}
-
-static bool remove_udf_in_table(const char *exact_name_str,
-                                uint exact_name_len,
-                                TABLE *table)
-{
-  if (find_udf_in_table(exact_name_str, exact_name_len, table))
-  {
-    int error;
-    if ((error= table->file->ha_delete_row(table->record[0])))
-    {
-      table->file->print_error(error, MYF(0));
-      return TRUE;
-    }
-  }
-  return FALSE;
-}
-
-
 /*
   Drop user defined function.
 
@@ -474,8 +431,8 @@ static int mysql_drop_function_internal(THD *thd, udf_func *udf, TABLE *table)
 {
   DBUG_ENTER("mysql_drop_function_internal");
 
-  char *exact_name_str= udf->name.str;
-  uint exact_name_len= udf->name.length;
+  const char *exact_name_str= udf->name.str;
+  size_t exact_name_len= udf->name.length;
 
   del_udf(udf);
   /*
@@ -488,21 +445,18 @@ static int mysql_drop_function_internal(THD *thd, udf_func *udf, TABLE *table)
   if (!table)
     DBUG_RETURN(1);
 
-  bool ret= remove_udf_in_table(exact_name_str, exact_name_len, table);
-  DBUG_RETURN(ret);
-}
-
-
-static TABLE *open_udf_func_table(THD *thd)
-{
-  TABLE_LIST tables;
-  TABLE *table;
-
-  tables.init_one_table(STRING_WITH_LEN("mysql"), STRING_WITH_LEN("func"),
-                        "func", TL_WRITE);
-  table= open_ltable(thd, &tables, TL_WRITE, MYSQL_LOCK_IGNORE_TIMEOUT);
-
-  return table;
+  table->use_all_columns();
+  table->field[0]->store(exact_name_str, exact_name_len, &my_charset_bin);
+  if (!table->file->ha_index_read_idx_map(table->record[0], 0,
+                                          (uchar*) table->field[0]->ptr,
+                                          HA_WHOLE_KEY,
+                                          HA_READ_KEY_EXACT))
+  {
+    int error;
+    if (unlikely((error= table->file->ha_delete_row(table->record[0]))))
+      table->file->print_error(error, MYF(0));
+  }
+  DBUG_RETURN(0);
 }
 
 
@@ -549,7 +503,8 @@ int mysql_create_function(THD *thd,udf_func *udf)
   if (check_ident_length(&udf->name))
     DBUG_RETURN(1);
 
-  table= open_udf_func_table(thd);
+  tables.init_one_table(&MYSQL_SCHEMA_NAME, &MYSQL_FUNC_NAME, 0, TL_WRITE);
+  table= open_ltable(thd, &tables, TL_WRITE, MYSQL_LOCK_IGNORE_TIMEOUT);
 
   mysql_rwlock_wrlock(&THR_LOCK_udf);
   DEBUG_SYNC(current_thd, "mysql_create_function_after_lock");
@@ -558,7 +513,7 @@ int mysql_create_function(THD *thd,udf_func *udf)
   {
     if (thd->lex->create_info.or_replace())
     {
-      if ((error= mysql_drop_function_internal(thd, u_d, table)))
+      if (unlikely((error= mysql_drop_function_internal(thd, u_d, table))))
         goto err;
     }
     else if (thd->lex->create_info.if_not_exists())
@@ -592,7 +547,8 @@ int mysql_create_function(THD *thd,udf_func *udf)
   }
   udf->dlhandle=dl;
   {
-    char buf[SAFE_NAME_LEN+16], *missing;
+    char buf[SAFE_NAME_LEN+16];
+    const char *missing;
     if ((missing= init_syms(udf, buf)))
     {
       my_error(ER_CANT_FIND_DL_ENTRY, MYF(0), missing);
@@ -613,7 +569,7 @@ int mysql_create_function(THD *thd,udf_func *udf)
   /* create entry in mysql.func table */
 
   /* Allow creation of functions even if we can't open func table */
-  if (!table)
+  if (unlikely(!table))
     goto err;
   table->use_all_columns();
   restore_record(table, s->default_values);	// Default values for fields
@@ -622,9 +578,9 @@ int mysql_create_function(THD *thd,udf_func *udf)
   table->field[2]->store(u_d->dl,(uint) strlen(u_d->dl), system_charset_info);
   if (table->s->fields >= 4)			// If not old func format
     table->field[3]->store((longlong) u_d->type, TRUE);
-  error = table->file->ha_write_row(table->record[0]);
+  error= table->file->ha_write_row(table->record[0]);
 
-  if (error)
+  if (unlikely(error))
   {
     my_error(ER_ERROR_ON_WRITE, MYF(0), "mysql.func", error);
     del_udf(u_d);
@@ -635,7 +591,7 @@ done:
   mysql_rwlock_unlock(&THR_LOCK_udf);
 
   /* Binlog the create function. */
-  if (write_bin_log(thd, TRUE, thd->query(), thd->query_length()))
+  if (unlikely(write_bin_log(thd, TRUE, thd->query(), thd->query_length())))
     DBUG_RETURN(1);
 
   DBUG_RETURN(0);
@@ -648,65 +604,42 @@ err:
 }
 
 
-enum drop_udf_result mysql_drop_function(THD *thd, const LEX_STRING *udf_name)
+int mysql_drop_function(THD *thd, const LEX_CSTRING *udf_name)
 {
   TABLE *table;
+  TABLE_LIST tables;
   udf_func *udf;
   DBUG_ENTER("mysql_drop_function");
 
-  if (thd->locked_tables_mode)
-  {
-    my_error(ER_LOCK_OR_ACTIVE_TRANSACTION, MYF(0));
-    DBUG_RETURN(UDF_DEL_RESULT_ERROR);
-  }
-
-  if (!(table= open_udf_func_table(thd)))
-    DBUG_RETURN(UDF_DEL_RESULT_ERROR);
-
-  // Fast pre-check
-  if (!mysql_rwlock_tryrdlock(&THR_LOCK_udf))
-  {
-    bool found=find_udf_everywhere(thd, udf_name->str, udf_name->length, table);
-    mysql_rwlock_unlock(&THR_LOCK_udf);
-    if (!found)
-    {
-      close_mysql_tables(thd);
-      DBUG_RETURN(UDF_DEL_RESULT_ABSENT);
-    }
-  }
-
   if (!initialized)
   {
-    close_mysql_tables(thd);
     if (opt_noacl)
-      DBUG_RETURN(UDF_DEL_RESULT_ABSENT); // SP should be checked
-
-    my_message(ER_OUT_OF_RESOURCES, ER_THD(thd, ER_OUT_OF_RESOURCES), MYF(0));
-    DBUG_RETURN(UDF_DEL_RESULT_ERROR);
+      my_error(ER_FUNCTION_NOT_DEFINED, MYF(0), udf_name->str);
+    else
+      my_message(ER_OUT_OF_RESOURCES, ER_THD(thd, ER_OUT_OF_RESOURCES),
+                 MYF(0));
+    DBUG_RETURN(1);
   }
 
+  tables.init_one_table(&MYSQL_SCHEMA_NAME, &MYSQL_FUNC_NAME, 0, TL_WRITE);
+  table= open_ltable(thd, &tables, TL_WRITE, MYSQL_LOCK_IGNORE_TIMEOUT);
+
   mysql_rwlock_wrlock(&THR_LOCK_udf);
-
-  // re-check under protection
-  if (!find_udf_everywhere(thd, udf_name->str, udf_name->length, table))
-  {
-    close_mysql_tables(thd);
-    mysql_rwlock_unlock(&THR_LOCK_udf);
-    DBUG_RETURN(UDF_DEL_RESULT_ABSENT);
-  }
-
-  if (check_access(thd, DELETE_ACL, "mysql", NULL, NULL, 1, 0))
-    goto err;
-
-
   DEBUG_SYNC(current_thd, "mysql_drop_function_after_lock");
-
   if (!(udf= (udf_func*) my_hash_search(&udf_hash, (uchar*) udf_name->str,
                                         (uint) udf_name->length)) )
   {
-    if (remove_udf_in_table(udf_name->str, (uint) udf_name->length, table))
-      goto err;
-    goto done;
+    if (thd->lex->check_exists)
+    {
+      push_warning_printf(thd, Sql_condition::WARN_LEVEL_NOTE,
+                          ER_FUNCTION_NOT_DEFINED,
+                          ER_THD(thd, ER_FUNCTION_NOT_DEFINED),
+                          udf_name->str);
+      goto done;
+    }
+
+    my_error(ER_FUNCTION_NOT_DEFINED, MYF(0), udf_name->str);
+    goto err;
   }
 
   if (mysql_drop_function_internal(thd, udf, table))
@@ -720,24 +653,13 @@ done:
     while binlogging, to avoid binlog inconsistency.
   */
   if (write_bin_log(thd, TRUE, thd->query(), thd->query_length()))
-    DBUG_RETURN(UDF_DEL_RESULT_ERROR);
+    DBUG_RETURN(1);
 
-  close_mysql_tables(thd);
-  DBUG_RETURN(UDF_DEL_RESULT_DELETED);
+  DBUG_RETURN(0);
 
 err:
-  close_mysql_tables(thd);
   mysql_rwlock_unlock(&THR_LOCK_udf);
-  DBUG_RETURN(UDF_DEL_RESULT_ERROR);
-}
-
-static bool find_udf_everywhere(THD* thd, const char *name, uint len,
-                                TABLE *table)
-{
-  if (initialized && my_hash_search(&udf_hash, (uchar*) name, len))
-    return true;
-
-  return find_udf_in_table(name, len, table);
+  DBUG_RETURN(1);
 }
 
 #endif /* HAVE_DLOPEN */
