@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 2007, 2018, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2016, 2020, MariaDB Corporation.
+Copyright (c) 2016, 2019, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -525,7 +525,7 @@ fts_index_fetch_nodes(
 	for (;;) {
 		error = fts_eval_sql(trx, *graph);
 
-		if (UNIV_LIKELY(error == DB_SUCCESS)) {
+		if (error == DB_SUCCESS) {
 			fts_sql_commit(trx);
 
 			break;				/* Exit the loop. */
@@ -538,7 +538,7 @@ fts_index_fetch_nodes(
 
 				trx->error_state = DB_SUCCESS;
 			} else {
-				ib::error() << "(" << error
+				ib::error() << "(" << ut_strerr(error)
 					<< ") while reading FTS index.";
 
 				break;			/* Exit the loop. */
@@ -615,9 +615,9 @@ fts_zip_read_word(
 				ptr[len] = 0;
 
 				zip->zp->next_out = ptr;
-				zip->zp->avail_out = len;
+				zip->zp->avail_out = uInt(len);
 
-				word->f_len = len;
+				word->f_len = ulint(len);
 				len = 0;
 			}
 			break;
@@ -709,7 +709,7 @@ fts_fetch_index_words(
 		case Z_OK:
 			if (zip->zp->avail_in == 0) {
 				zip->zp->next_in = static_cast<byte*>(data);
-				zip->zp->avail_in = len;
+				zip->zp->avail_in = uInt(len);
 				ut_a(len <= FTS_MAX_WORD_LEN);
 				len = 0;
 			}
@@ -860,7 +860,7 @@ fts_index_fetch_words(
 				error = fts_eval_sql(optim->trx, graph);
 			}
 
-			if (UNIV_LIKELY(error == DB_SUCCESS)) {
+			if (error == DB_SUCCESS) {
 				//FIXME fts_sql_commit(optim->trx);
 				break;
 			} else {
@@ -877,7 +877,7 @@ fts_index_fetch_words(
 
 					optim->trx->error_state = DB_SUCCESS;
 				} else {
-					ib::error() << "(" << error
+					ib::error() << "(" << ut_strerr(error)
 						<< ") while reading document.";
 
 					break;	/* Exit the loop. */
@@ -920,7 +920,7 @@ fts_fetch_doc_ids(
 	int		i = 0;
 	sel_node_t*	sel_node = static_cast<sel_node_t*>(row);
 	fts_doc_ids_t*	fts_doc_ids = static_cast<fts_doc_ids_t*>(user_arg);
-	doc_id_t*	update = static_cast<doc_id_t*>(
+	fts_update_t*	update = static_cast<fts_update_t*>(
 		ib_vector_push(fts_doc_ids->doc_ids, NULL));
 
 	for (exp = sel_node->select_list;
@@ -936,7 +936,8 @@ fts_fetch_doc_ids(
 		/* Note: The column numbers below must match the SELECT. */
 		switch (i) {
 		case 0: /* DOC_ID */
-			*update = fts_read_doc_id(
+			update->fts_indexes = NULL;
+			update->doc_id = fts_read_doc_id(
 				static_cast<byte*>(data));
 			break;
 
@@ -968,7 +969,7 @@ fts_table_fetch_doc_ids(
 	ut_a(fts_table->type == FTS_COMMON_TABLE);
 
 	if (!trx) {
-		trx = trx_allocate_for_background();
+		trx = trx_create();
 		alloc_bk_trx = TRUE;
 	}
 
@@ -1004,11 +1005,11 @@ fts_table_fetch_doc_ids(
 	mutex_exit(&dict_sys->mutex);
 
 	if (error == DB_SUCCESS) {
-		ib_vector_sort(doc_ids->doc_ids, fts_doc_id_cmp);
+		ib_vector_sort(doc_ids->doc_ids, fts_update_doc_id_cmp);
 	}
 
 	if (alloc_bk_trx) {
-		trx_free_for_background(trx);
+		trx_free(trx);
 	}
 
 	return(error);
@@ -1021,7 +1022,7 @@ Do a binary search for a doc id in the array
 int
 fts_bsearch(
 /*========*/
-	doc_id_t*	array,	/*!< in: array to sort */
+	fts_update_t*	array,	/*!< in: array to sort */
 	int		lower,	/*!< in: the array lower bound */
 	int		upper,	/*!< in: the array upper bound */
 	doc_id_t	doc_id)	/*!< in: the doc id to search for */
@@ -1035,9 +1036,9 @@ fts_bsearch(
 		while (lower < upper) {
 			int	i = (lower + upper) >> 1;
 
-			if (doc_id > array[i]) {
+			if (doc_id > array[i].doc_id) {
 				lower = i + 1;
-			} else if (doc_id < array[i]) {
+			} else if (doc_id < array[i].doc_id) {
 				upper = i - 1;
 			} else {
 				return(i); /* Found. */
@@ -1046,7 +1047,7 @@ fts_bsearch(
 	}
 
 	if (lower == upper && lower < orig_size) {
-		if (doc_id == array[lower]) {
+		if (doc_id == array[lower].doc_id) {
 			return(lower);
 		} else if (lower == 0) {
 			return(-1);
@@ -1073,7 +1074,7 @@ fts_optimize_lookup(
 {
 	int		pos;
 	int		upper = static_cast<int>(ib_vector_size(doc_ids));
-	doc_id_t*	array = (doc_id_t*) doc_ids->data;
+	fts_update_t*	array = (fts_update_t*) doc_ids->data;
 
 	pos = fts_bsearch(array, static_cast<int>(lower), upper, first_doc_id);
 
@@ -1086,10 +1087,10 @@ fts_optimize_lookup(
 		/* If i is 1, it could be first_doc_id is less than
 		either the first or second array item, do a
 		double check */
-		if (i == 1 && array[0] <= last_doc_id
-		    && first_doc_id < array[0]) {
+		if (i == 1 && array[0].doc_id <= last_doc_id
+		    && first_doc_id < array[0].doc_id) {
 			pos = 0;
-		} else if (i < upper && array[i] <= last_doc_id) {
+		} else if (i < upper && array[i].doc_id <= last_doc_id) {
 
 			/* Check if the "next" doc id is within the
 			first & last doc id of the node. */
@@ -1138,7 +1139,7 @@ fts_optimize_encode_node(
 	++src;
 
 	/* Number of encoded pos bytes to copy. */
-	pos_enc_len = src - enc->src_ilist_ptr;
+	pos_enc_len = ulint(src - enc->src_ilist_ptr);
 
 	/* Total number of bytes required for copy. */
 	enc_len += pos_enc_len;
@@ -1210,7 +1211,7 @@ fts_optimize_node(
 		enc->src_ilist_ptr = src_node->ilist;
 	}
 
-	copied = enc->src_ilist_ptr - src_node->ilist;
+	copied = ulint(enc->src_ilist_ptr - src_node->ilist);
 
 	/* While there is data in the source node and space to copy
 	into in the destination node. */
@@ -1228,12 +1229,12 @@ test_again:
 		delta for decoding the entries following this document's
 		entries. */
 		if (*del_pos >= 0 && *del_pos < (int) ib_vector_size(del_vec)) {
-			doc_id_t*	update;
+			fts_update_t*	update;
 
-			update = (doc_id_t*) ib_vector_get(
-				del_vec, *del_pos);
+			update = (fts_update_t*) ib_vector_get(
+				del_vec, ulint(*del_pos));
 
-			del_doc_id = *update;
+			del_doc_id = update->doc_id;
 		}
 
 		if (enc->src_ilist_ptr == src_node->ilist && doc_id == 0) {
@@ -1275,7 +1276,7 @@ test_again:
 		}
 
 		/* Bytes copied so for from source. */
-		copied = enc->src_ilist_ptr - src_node->ilist;
+		copied = ulint(enc->src_ilist_ptr - src_node->ilist);
 	}
 
 	if (copied >= src_node->ilist_size) {
@@ -1356,6 +1357,12 @@ fts_optimize_word(
 	enc.src_last_doc_id = 0;
 	enc.src_ilist_ptr = NULL;
 
+	if (fts_enable_diag_print) {
+		word->text.f_str[word->text.f_len] = 0;
+		ib::info() << "FTS_OPTIMIZE: optimize \"" << word->text.f_str
+			<< "\"";
+	}
+
 	while (i < size) {
 		ulint		copied;
 		fts_node_t*	src_node;
@@ -1376,7 +1383,7 @@ fts_optimize_word(
 		ut_a(enc.src_ilist_ptr != NULL);
 
 		/* Determine the numer of bytes copied to dst_node. */
-		copied = enc.src_ilist_ptr - src_node->ilist;
+		copied = ulint(enc.src_ilist_ptr - src_node->ilist);
 
 		/* Can't copy more than whats in the vlc array. */
 		ut_a(copied <= src_node->ilist_size);
@@ -1433,6 +1440,11 @@ fts_optimize_write_word(
 
 	ut_ad(fts_table->charset);
 
+	if (fts_enable_diag_print) {
+		ib::info() << "FTS_OPTIMIZE: processed \"" << word->f_str
+			<< "\"";
+	}
+
 	pars_info_bind_varchar_literal(
 		info, "word", word->f_str, word->f_len);
 
@@ -1450,8 +1462,8 @@ fts_optimize_write_word(
 
 	error = fts_eval_sql(trx, graph);
 
-	if (UNIV_UNLIKELY(error != DB_SUCCESS)) {
-		ib::error() << "(" << error << ") during optimize,"
+	if (error != DB_SUCCESS) {
+		ib::error() << "(" << ut_strerr(error) << ") during optimize,"
 			" when deleting a word from the FTS index.";
 	}
 
@@ -1474,8 +1486,8 @@ fts_optimize_write_word(
 			error = fts_write_node(
 				trx, &graph, fts_table, word, node);
 
-			if (UNIV_UNLIKELY(error != DB_SUCCESS)) {
-				ib::error() << "(" << error << ")"
+			if (error != DB_SUCCESS) {
+				ib::error() << "(" << ut_strerr(error) << ")"
 					" during optimize, while adding a"
 					" word to the FTS index.";
 			}
@@ -1587,7 +1599,7 @@ fts_optimize_create(
 
 	optim->table = table;
 
-	optim->trx = trx_allocate_for_background();
+	optim->trx = trx_create();
 	trx_start_internal(optim->trx);
 
 	optim->fts_common_table.table_id = table->id;
@@ -1710,7 +1722,7 @@ fts_optimize_free(
 	mem_heap_t*	heap = static_cast<mem_heap_t*>(optim->self_heap->arg);
 
 	trx_commit_for_mysql(optim->trx);
-	trx_free_for_background(optim->trx);
+	trx_free(optim->trx);
 
 	fts_doc_ids_free(optim->to_delete);
 	fts_optimize_graph_free(&optim->graph);
@@ -1864,8 +1876,9 @@ fts_optimize_index_completed(
 	error = fts_config_set_index_value(
 		optim->trx, index, FTS_LAST_OPTIMIZED_WORD, &word);
 
-	if (UNIV_UNLIKELY(error != DB_SUCCESS)) {
-		ib::error() << "(" << error << ") while updating"
+	if (error != DB_SUCCESS) {
+
+		ib::error() << "(" << ut_strerr(error) << ") while updating"
 			" last optimized word!";
 	}
 
@@ -2007,7 +2020,7 @@ fts_optimize_purge_deleted_doc_ids(
 	ulint		i;
 	pars_info_t*	info;
 	que_t*		graph;
-	doc_id_t*	update;
+	fts_update_t*	update;
 	doc_id_t	write_doc_id;
 	dberr_t		error = DB_SUCCESS;
 	char		deleted[MAX_FULL_NAME_LEN];
@@ -2017,11 +2030,11 @@ fts_optimize_purge_deleted_doc_ids(
 
 	ut_a(ib_vector_size(optim->to_delete->doc_ids) > 0);
 
-	update = static_cast<doc_id_t*>(
+	update = static_cast<fts_update_t*>(
 		ib_vector_get(optim->to_delete->doc_ids, 0));
 
 	/* Convert to "storage" byte order. */
-	fts_write_doc_id((byte*) &write_doc_id, *update);
+	fts_write_doc_id((byte*) &write_doc_id, update->doc_id);
 
 	/* This is required for the SQL parser to work. It must be able
 	to find the following variables. So we do it twice. */
@@ -2043,11 +2056,11 @@ fts_optimize_purge_deleted_doc_ids(
 	/* Delete the doc ids that were copied at the start. */
 	for (i = 0; i < ib_vector_size(optim->to_delete->doc_ids); ++i) {
 
-		update = static_cast<doc_id_t*>(ib_vector_get(
+		update = static_cast<fts_update_t*>(ib_vector_get(
 			optim->to_delete->doc_ids, i));
 
 		/* Convert to "storage" byte order. */
-		fts_write_doc_id((byte*) &write_doc_id, *update);
+		fts_write_doc_id((byte*) &write_doc_id, update->doc_id);
 
 		fts_bind_doc_id(info, "doc_id1", &write_doc_id);
 
@@ -2417,7 +2430,7 @@ fts_optimize_table(
 	fts_optimize_t*	optim = NULL;
 	fts_t*		fts = table->fts;
 
-	if (UNIV_UNLIKELY(fts_enable_diag_print)) {
+	if (fts_enable_diag_print) {
 		ib::info() << "FTS start optimize " << table->name;
 	}
 
@@ -2469,7 +2482,7 @@ fts_optimize_table(
 		if (error == DB_SUCCESS
 		    && optim->n_completed == ib_vector_size(fts->indexes)) {
 
-			if (UNIV_UNLIKELY(fts_enable_diag_print)) {
+			if (fts_enable_diag_print) {
 				ib::info() << "FTS_OPTIMIZE: Completed"
 					" Optimize, cleanup DELETED table";
 			}
@@ -2492,7 +2505,7 @@ fts_optimize_table(
 
 	fts_optimize_free(optim);
 
-	if (UNIV_UNLIKELY(fts_enable_diag_print)) {
+	if (fts_enable_diag_print) {
 		ib::info() << "FTS end optimize " << table->name;
 	}
 
@@ -2677,7 +2690,7 @@ static bool fts_optimize_del_table(const dict_table_t* table)
 		slot = static_cast<fts_slot_t*>(ib_vector_get(fts_slots, i));
 
 		if (slot->table == table) {
-			if (UNIV_UNLIKELY(fts_enable_diag_print)) {
+			if (fts_enable_diag_print) {
 				ib::info() << "FTS Optimize Removing table "
 					<< table->name;
 			}
@@ -2772,7 +2785,7 @@ Optimize all FTS tables.
 @return Dummy return */
 static
 os_thread_ret_t
-fts_optimize_thread(
+DECLARE_THREAD(fts_optimize_thread)(
 /*================*/
 	void*		arg)			/*!< in: work queue*/
 {
@@ -2790,7 +2803,8 @@ fts_optimize_thread(
 	/* Assign number of tables added in fts_slots_t to n_tables */
 	n_tables = ib_vector_size(fts_slots);
 
-	while (!done && srv_shutdown_state <= SRV_SHUTDOWN_INITIATED) {
+	while (!done && srv_shutdown_state == SRV_SHUTDOWN_NONE) {
+
 		/* If there is no message in the queue and we have tables
 		to optimize then optimize the tables. */
 
