@@ -39,7 +39,6 @@ Created 2/6/1997 Heikki Tuuri
 #include "row0row.h"
 #include "row0upd.h"
 #include "rem0cmp.h"
-#include "read0read.h"
 #include "lock0lock.h"
 #include "row0mysql.h"
 
@@ -59,7 +58,7 @@ row_vers_non_virtual_fields_equal(
 
 	for (const dict_field_t* ifield = index->fields; ifield != end;
 	     ifield++) {
-		if (!dict_col_is_virtual(ifield->col)
+		if (!ifield->col->is_virtual()
 		    && cmp_dfield_dfield(a++, b++)) {
 			return false;
 		}
@@ -70,6 +69,7 @@ row_vers_non_virtual_fields_equal(
 
 /** Determine if an active transaction has inserted or modified a secondary
 index record.
+@param[in,out]	caller_trx	trx of current thread
 @param[in]	clust_rec	clustered index record
 @param[in]	clust_index	clustered index
 @param[in]	rec		secondary index record
@@ -82,6 +82,7 @@ trx_mutex_enter(), and trx->release_reference() must be invoked
 UNIV_INLINE
 trx_t*
 row_vers_impl_x_locked_low(
+	trx_t*		caller_trx,
 	const rec_t*	clust_rec,
 	dict_index_t*	clust_index,
 	const rec_t*	rec,
@@ -89,9 +90,10 @@ row_vers_impl_x_locked_low(
 	const rec_offs*	offsets,
 	mtr_t*		mtr)
 {
+	trx_id_t	trx_id;
 	rec_t*		prev_version = NULL;
 	rec_offs	clust_offsets_[REC_OFFS_NORMAL_SIZE];
-	rec_offs*	clust_offsets;
+	rec_offs*	clust_offsets = clust_offsets_;
 	mem_heap_t*	heap;
 	dtuple_t*	ientry = NULL;
 	mem_heap_t*	v_heap = NULL;
@@ -103,29 +105,44 @@ row_vers_impl_x_locked_low(
 
 	ut_ad(rec_offs_validate(rec, index, offsets));
 
+	if (ulint trx_id_offset = clust_index->trx_id_offset) {
+		trx_id = mach_read_from_6(clust_rec + trx_id_offset);
+		if (trx_id == 0) {
+			/* The transaction history was already purged. */
+			DBUG_RETURN(0);
+		}
+	}
+
 	heap = mem_heap_create(1024);
 
-	clust_offsets = rec_get_offsets(clust_rec, clust_index, clust_offsets_,
+	clust_offsets = rec_get_offsets(clust_rec, clust_index, clust_offsets,
 					true, ULINT_UNDEFINED, &heap);
 
-	const trx_id_t trx_id = row_get_rec_trx_id(
-		clust_rec, clust_index, clust_offsets);
+	trx_id = row_get_rec_trx_id(clust_rec, clust_index, clust_offsets);
+	if (trx_id == 0) {
+		/* The transaction history was already purged. */
+		mem_heap_free(heap);
+		DBUG_RETURN(0);
+	}
 
 	ut_ad(!clust_index->table->is_temporary());
 
-	bool corrupt = false;
-	trx_t*	trx = trx_rw_is_active(trx_id, &corrupt, true);
+	trx_t*	trx;
 
-	if (trx == 0) {
-		/* The transaction that modified or inserted clust_rec is no
-		longer active, or it is corrupt: no implicit lock on rec */
-		if (UNIV_UNLIKELY(corrupt)) {
-			lock_report_trx_id_insanity(
-				trx_id, clust_rec, clust_index, clust_offsets,
-				trx_sys_get_max_trx_id());
+	if (trx_id == caller_trx->id) {
+		trx = caller_trx;
+		trx->reference();
+	} else {
+		trx = trx_sys.find(caller_trx, trx_id);
+		if (trx == 0) {
+			/* The transaction that modified or inserted
+			clust_rec is no longer active, or it is
+			corrupt: no implicit lock on rec */
+			lock_check_trx_id_sanity(trx_id, clust_rec,
+						 clust_index, clust_offsets);
+			mem_heap_free(heap);
+			DBUG_RETURN(0);
 		}
-		mem_heap_free(heap);
-		DBUG_RETURN(0);
 	}
 
 	const ulint comp = page_rec_is_comp(rec);
@@ -187,7 +204,7 @@ row_vers_impl_x_locked_low(
 		ut_ad(committed || prev_version
 		      || !rec_get_deleted_flag(version, comp));
 
-		/* Free version and clust_offsets. */
+		/* Free version. */
 		mem_heap_free(old_heap);
 
 		if (committed) {
@@ -222,7 +239,7 @@ not_locked:
 		}
 
 		clust_offsets = rec_get_offsets(
-			prev_version, clust_index, clust_offsets_, true,
+			prev_version, clust_index, clust_offsets, true,
 			ULINT_UNDEFINED, &heap);
 
 		vers_del = rec_get_deleted_flag(prev_version, comp);
@@ -359,6 +376,7 @@ result_check:
 
 /** Determine if an active transaction has inserted or modified a secondary
 index record.
+@param[in,out]	caller_trx	trx of current thread
 @param[in]	rec	secondary index record
 @param[in]	index	secondary index
 @param[in]	offsets	rec_get_offsets(rec, index)
@@ -367,6 +385,7 @@ trx_mutex_enter(), and trx->release_reference() must be invoked
 @retval	NULL if the record was committed */
 trx_t*
 row_vers_impl_x_locked(
+	trx_t*		caller_trx,
 	const rec_t*	rec,
 	dict_index_t*	index,
 	const rec_offs*	offsets)
@@ -377,7 +396,7 @@ row_vers_impl_x_locked(
 	dict_index_t*	clust_index;
 
 	ut_ad(!lock_mutex_own());
-	ut_ad(!trx_sys_mutex_own());
+	ut_ad(!mutex_own(&trx_sys.mutex));
 
 	mtr_start(&mtr);
 
@@ -407,7 +426,8 @@ row_vers_impl_x_locked(
 		trx = 0;
 	} else {
 		trx = row_vers_impl_x_locked_low(
-			clust_rec, clust_index, rec, index, offsets, &mtr);
+				caller_trx, clust_rec, clust_index, rec, index,
+				offsets, &mtr);
 
 		ut_ad(trx == 0 || trx->is_referenced());
 	}
@@ -417,39 +437,14 @@ row_vers_impl_x_locked(
 	return(trx);
 }
 
-/*****************************************************************//**
-Finds out if we must preserve a delete marked earlier version of a clustered
-index record, because it is >= the purge view.
-@param[in]	trx_id		transaction id in the version
-@param[in]	name		table name
-@param[in,out]	mtr		mini transaction holding the latch on the
-				clustered index record; it will also hold
-				the latch on purge_view
-@return TRUE if earlier version should be preserved */
-ibool
-row_vers_must_preserve_del_marked(
-/*==============================*/
-	trx_id_t		trx_id,
-	const table_name_t&	name,
-	mtr_t*			mtr)
-{
-	ut_ad(!rw_lock_own(&(purge_sys->latch), RW_LOCK_S));
-
-	mtr_s_lock(&purge_sys->latch, mtr);
-
-	return(!purge_sys->view.changes_visible(trx_id,	name));
-}
-
 /** build virtual column value from current cluster index record data
 @param[in,out]	row		the cluster index row in dtuple form
 @param[in]	clust_index	clustered index
 @param[in]	index		the secondary index
 @param[in]	heap		heap used to build virtual dtuple
-@param[in,out]	vcol_info	virtual column information.
-@return		true in case of success
-		false if virtual column computation fails */
+@param[in,out]	vcol_info	virtual column information. */
 static
-bool
+void
 row_vers_build_clust_v_col(
 	dtuple_t*		row,
 	dict_index_t*		clust_index,
@@ -457,8 +452,11 @@ row_vers_build_clust_v_col(
 	mem_heap_t*		heap,
 	purge_vcol_info_t*	vcol_info)
 {
+	mem_heap_t*	local_heap = NULL;
+	VCOL_STORAGE	*vcol_storage= NULL;
 	THD*		thd= current_thd;
 	TABLE*		maria_table= 0;
+	byte*		record= 0;
 
 	ut_ad(dict_index_has_virtual(index));
 	ut_ad(index->table == clust_index->table);
@@ -469,37 +467,40 @@ row_vers_build_clust_v_col(
 	}
 	DEBUG_SYNC(current_thd, "ib_clust_v_col_before_row_allocated");
 
-	ib_vcol_row vc(NULL);
-	byte *record = vc.record(thd, index, &maria_table);
+	innobase_allocate_row_for_vcol(thd, index,
+				       &local_heap,
+				       &maria_table,
+				       &record,
+				       &vcol_storage);
 
 	if (vcol_info && !vcol_info->table()) {
 		vcol_info->set_table(maria_table);
-		// wait for second fetch
-		return true;
+		goto func_exit;
 	}
 
 	for (ulint i = 0; i < dict_index_get_n_fields(index); i++) {
 		const dict_field_t* ind_field = dict_index_get_nth_field(
 				index, i);
 
-		if (dict_col_is_virtual(ind_field->col)) {
+		if (ind_field->col->is_virtual()) {
 			const dict_v_col_t*       col;
 
 			col = reinterpret_cast<const dict_v_col_t*>(
 				ind_field->col);
 
-			dfield_t *vfield = innobase_get_computed_value(
-				row, col, clust_index, &vc.heap,
+			innobase_get_computed_value(
+				row, col, clust_index, &local_heap,
 				heap, NULL, thd, maria_table, record, NULL,
-				NULL);
-			if (vfield == NULL) {
-				innobase_report_computed_value_failed(row);
-				ut_ad(0);
-				return false;
-			}
+				NULL, NULL);
 		}
 	}
-	return true;
+
+func_exit:
+	if (local_heap) {
+		if (vcol_storage)
+			innobase_free_row_for_vcol(vcol_storage);
+		mem_heap_free(local_heap);
+	}
 }
 
 /** Build latest virtual column data from undo log
@@ -584,7 +585,7 @@ row_vers_build_cur_vrow_low(
 				 = dict_index_get_nth_field(index, i);
 			const dict_col_t*	col = ind_field->col;
 
-			if (!dict_col_is_virtual(col)) {
+			if (!col->is_virtual()) {
 				continue;
 			}
 
@@ -666,7 +667,7 @@ row_vers_vc_matches_cluster(
 		for (const dict_field_t *ifield = index->fields,
 			     *const end = &index->fields[index->n_fields];
 		     ifield != end; ifield++, a++, b++) {
-			if (!dict_col_is_virtual(ifield->col)) {
+			if (!ifield->col->is_virtual()) {
 				if (cmp_dfield_dfield(a, b)) {
 					return false;
 				}
@@ -725,7 +726,7 @@ row_vers_vc_matches_cluster(
 			const dict_col_t*	col = ind_field->col;
 			field1 = dtuple_get_nth_field(ientry, i);
 
-			if (!dict_col_is_virtual(col)) {
+			if (!col->is_virtual()) {
 				continue;
 			}
 
@@ -790,7 +791,6 @@ func_exit:
 @param[in]	clust_index	cluster index
 @param[in]	clust_offsets	cluster rec offset
 @param[in]	index		secondary index
-@param[in]	ientry		secondary index rec
 @param[in]	roll_ptr	roll_ptr for the purge record
 @param[in]	trx_id		transaction ID on the purging record
 @param[in,out]	heap		heap memory
@@ -806,7 +806,6 @@ row_vers_build_cur_vrow(
 	dict_index_t*		clust_index,
 	rec_offs**		clust_offsets,
 	dict_index_t*		index,
-	const dtuple_t*		ientry,
 	roll_ptr_t		roll_ptr,
 	trx_id_t		trx_id,
 	mem_heap_t*		heap,
@@ -836,11 +835,8 @@ row_vers_build_cur_vrow(
 			mtr->commit();
 		}
 
-		bool res = row_vers_build_clust_v_col(
+		row_vers_build_clust_v_col(
 			row, clust_index, index, heap, vcol_info);
-		if (!res) {
-			return NULL;
-		}
 
 		if (vcol_info != NULL && vcol_info->is_first_fetch()) {
 			return NULL;
@@ -904,7 +900,7 @@ row_vers_old_has_index_entry(
 
 	ut_ad(mtr_memo_contains_page_flagged(mtr, rec, MTR_MEMO_PAGE_X_FIX
 					     | MTR_MEMO_PAGE_S_FIX));
-	ut_ad(!rw_lock_own(&(purge_sys->latch), RW_LOCK_S));
+	ut_ad(!rw_lock_own(&purge_sys.latch, RW_LOCK_S));
 	ut_ad(also_curr || !vcol_info);
 
 	clust_index = dict_table_get_first_index(index->table);
@@ -928,7 +924,7 @@ row_vers_old_has_index_entry(
 		/* The top of the stack of versions is locked by the
 		mtr holding a latch on the page containing the
 		clustered index record. The bottom of the stack is
-		locked by the fact that the purge_sys->view must
+		locked by the fact that the purge_sys.view must
 		'overtake' any read view of an active transaction.
 		Thus, it is safe to fetch the prefixes for
 		externally stored columns. */
@@ -961,14 +957,10 @@ row_vers_old_has_index_entry(
 					mtr->commit();
 				}
 
-				bool res = row_vers_build_clust_v_col(
+				row_vers_build_clust_v_col(
 					row, clust_index, index, heap,
 					vcol_info);
 
-				if (!res) {
-					goto unsafe_to_purge;
-				}
-
 				if (vcol_info && vcol_info->is_first_fetch()) {
 					goto unsafe_to_purge;
 				}
@@ -1047,8 +1039,7 @@ unsafe_to_purge:
 
 		cur_vrow = row_vers_build_cur_vrow(
 			also_curr, rec, clust_index, &clust_offsets,
-			index, ientry, roll_ptr, trx_id, heap, v_heap, mtr,
-			vcol_info);
+			index, roll_ptr, trx_id, heap, v_heap, mtr, vcol_info);
 
 		if (vcol_info && vcol_info->is_first_fetch()) {
 			goto unsafe_to_purge;
@@ -1179,7 +1170,7 @@ row_vers_build_for_consistent_read(
 	ut_ad(dict_index_is_clust(index));
 	ut_ad(mtr_memo_contains_page_flagged(mtr, rec, MTR_MEMO_PAGE_X_FIX
 					     | MTR_MEMO_PAGE_S_FIX));
-	ut_ad(!rw_lock_own(&(purge_sys->latch), RW_LOCK_S));
+	ut_ad(!rw_lock_own(&(purge_sys.latch), RW_LOCK_S));
 
 	ut_ad(rec_offs_validate(rec, index, *offsets));
 
@@ -1240,7 +1231,7 @@ row_vers_build_for_consistent_read(
 					in_heap, rec_offs_size(*offsets)));
 
 			*old_vers = rec_copy(buf, prev_version, *offsets);
-			rec_offs_make_valid(*old_vers, index, *offsets);
+			rec_offs_make_valid(*old_vers, index, true, *offsets);
 
 			if (vrow && *vrow) {
 				*vrow = dtuple_copy(*vrow, in_heap);
@@ -1263,6 +1254,7 @@ which should be seen by a semi-consistent read. */
 void
 row_vers_build_for_semi_consistent_read(
 /*====================================*/
+	trx_t*		caller_trx,/*!<in/out: trx of current thread */
 	const rec_t*	rec,	/*!< in: record in a clustered index; the
 				caller must have a latch on the page; this
 				latch locks the top of the stack of versions
@@ -1291,7 +1283,7 @@ row_vers_build_for_semi_consistent_read(
 	ut_ad(dict_index_is_clust(index));
 	ut_ad(mtr_memo_contains_page_flagged(mtr, rec, MTR_MEMO_PAGE_X_FIX
 					     | MTR_MEMO_PAGE_S_FIX));
-	ut_ad(!rw_lock_own(&(purge_sys->latch), RW_LOCK_S));
+	ut_ad(!rw_lock_own(&(purge_sys.latch), RW_LOCK_S));
 
 	ut_ad(rec_offs_validate(rec, index, *offsets));
 
@@ -1299,7 +1291,6 @@ row_vers_build_for_semi_consistent_read(
 	ut_ad(!vrow || !(*vrow));
 
 	for (;;) {
-		const trx_t*	version_trx;
 		mem_heap_t*	heap2;
 		rec_t*		prev_version;
 		trx_id_t	version_trx_id;
@@ -1309,20 +1300,7 @@ row_vers_build_for_semi_consistent_read(
 			rec_trx_id = version_trx_id;
 		}
 
-		trx_sys_mutex_enter();
-		version_trx = trx_get_rw_trx_by_id(version_trx_id);
-		/* Because version_trx is a read-write transaction,
-		its state cannot change from or to NOT_STARTED while
-		we are holding the trx_sys->mutex.  It may change from
-		ACTIVE to PREPARED or COMMITTED. */
-		if (version_trx
-		    && trx_state_eq(version_trx,
-				    TRX_STATE_COMMITTED_IN_MEMORY)) {
-			version_trx = NULL;
-		}
-		trx_sys_mutex_exit();
-
-		if (!version_trx) {
+		if (!trx_sys.is_registered(caller_trx, version_trx_id)) {
 committed_version_trx:
 			/* We found a version that belongs to a
 			committed transaction: return it. */
@@ -1363,7 +1341,7 @@ committed_version_trx:
 					in_heap, rec_offs_size(*offsets)));
 
 			*old_vers = rec_copy(buf, version, *offsets);
-			rec_offs_make_valid(*old_vers, index, *offsets);
+			rec_offs_make_valid(*old_vers, index, true, *offsets);
 			if (vrow && *vrow) {
 				*vrow = dtuple_copy(*vrow, in_heap);
 				dtuple_dup_v_fld(*vrow, in_heap);
