@@ -230,9 +230,7 @@ rw_lock_create_func(
 	ut_ad(cline <= 8192);
 	lock->cline = cline;
 	lock->count_os_wait = 0;
-	lock->last_s_file_name = "not yet reserved";
 	lock->last_x_file_name = "not yet reserved";
-	lock->last_s_line = 0;
 	lock->last_x_line = 0;
 	lock->event = os_event_create(0);
 	lock->wait_ex_event = os_event_create(0);
@@ -256,7 +254,8 @@ rw_lock_free_func(
 	rw_lock_t*	lock)	/*!< in/out: rw-lock */
 {
 	ut_ad(rw_lock_validate(lock));
-	ut_a(lock->lock_word == X_LOCK_DECR);
+	ut_a(my_atomic_load32_explicit(&lock->lock_word,
+				       MY_MEMORY_ORDER_RELAXED) == X_LOCK_DECR);
 
 	mutex_enter(&rw_lock_list_mutex);
 
@@ -285,22 +284,21 @@ rw_lock_s_lock_spin(
 {
 	ulint		i = 0;	/* spin round count */
 	sync_array_t*	sync_arr;
-	ulint		spin_count = 0;
-	uint64_t	count_os_wait = 0;
+	lint		spin_count = 0;
+	int64_t		count_os_wait = 0;
 
 	/* We reuse the thread id to index into the counter, cache
 	it here for efficiency. */
 
 	ut_ad(rw_lock_validate(lock));
 
-	rw_lock_stats.rw_s_spin_wait_count.inc();
-
 lock_loop:
 
 	/* Spin waiting for the writer field to become free */
 	HMT_low();
-	ulint j = i;
-	while (i < srv_n_spin_wait_rounds && lock->lock_word <= 0) {
+	while (i < srv_n_spin_wait_rounds &&
+	       my_atomic_load32_explicit(&lock->lock_word,
+					 MY_MEMORY_ORDER_RELAXED) <= 0) {
 		ut_delay(srv_spin_wait_delay);
 		i++;
 	}
@@ -310,7 +308,7 @@ lock_loop:
 		os_thread_yield();
 	}
 
-	spin_count += lint(i - j);
+	++spin_count;
 
 	/* We try once again to obtain the lock */
 	if (rw_lock_s_lock_low(lock, pass, file_name, line)) {
@@ -340,7 +338,7 @@ lock_loop:
 
 		/* Set waiters before checking lock_word to ensure wake-up
 		signal is sent. This may lead to some unnecessary signals. */
-		my_atomic_fas32((int32*) &lock->waiters, 1);
+		my_atomic_fas32_explicit(&lock->waiters, 1, MY_MEMORY_ORDER_ACQUIRE);
 
 		if (rw_lock_s_lock_low(lock, pass, file_name, line)) {
 
@@ -414,24 +412,23 @@ rw_lock_x_lock_wait_func(
 	unsigned	line)	/*!< in: line where requested */
 {
 	ulint		i = 0;
-	ulint		n_spins = 0;
+	lint		n_spins = 0;
 	sync_array_t*	sync_arr;
-	uint64_t	count_os_wait = 0;
+	int64_t		count_os_wait = 0;
 
-	ut_ad(lock->lock_word <= threshold);
+	ut_ad(my_atomic_load32_explicit(&lock->lock_word, MY_MEMORY_ORDER_RELAXED) <= threshold);
 
 	HMT_low();
-	while (lock->lock_word < threshold) {
+	while (my_atomic_load32_explicit(&lock->lock_word, MY_MEMORY_ORDER_RELAXED) < threshold) {
 		ut_delay(srv_spin_wait_delay);
 
 		if (i < srv_n_spin_wait_rounds) {
 			i++;
 			continue;
 		}
-		HMT_medium();
 
 		/* If there is still a reader, then go to sleep.*/
-		n_spins += i;
+		++n_spins;
 
 		sync_cell_t*	cell;
 
@@ -441,7 +438,7 @@ rw_lock_x_lock_wait_func(
 		i = 0;
 
 		/* Check lock_word to ensure wake-up isn't missed.*/
-		if (lock->lock_word < threshold) {
+		if (my_atomic_load32_explicit(&lock->lock_word, MY_MEMORY_ORDER_RELAXED) < threshold) {
 
 			++count_os_wait;
 
@@ -464,7 +461,6 @@ rw_lock_x_lock_wait_func(
 			sync_array_free_cell(sync_arr, cell);
 			break;
 		}
-		HMT_low();
 	}
 	HMT_medium();
 	rw_lock_stats.rw_x_spin_round_count.add(n_spins);
@@ -532,14 +528,18 @@ rw_lock_x_lock_low(
 					file_name, line);
 
 			} else {
+				int32_t lock_word = my_atomic_load32_explicit(&lock->lock_word,
+									      MY_MEMORY_ORDER_RELAXED);
 				/* At least one X lock by this thread already
 				exists. Add another. */
-				if (lock->lock_word == 0
-				    || lock->lock_word == -X_LOCK_HALF_DECR) {
-					lock->lock_word -= X_LOCK_DECR;
+				if (lock_word == 0
+				    || lock_word == -X_LOCK_HALF_DECR) {
+					my_atomic_add32_explicit(&lock->lock_word, -X_LOCK_DECR,
+								 MY_MEMORY_ORDER_RELAXED);
 				} else {
-					ut_ad(lock->lock_word <= -X_LOCK_DECR);
-					--lock->lock_word;
+					ut_ad(lock_word <= -X_LOCK_DECR);
+					my_atomic_add32_explicit(&lock->lock_word, -1,
+								 MY_MEMORY_ORDER_RELAXED);
 				}
 			}
 
@@ -610,12 +610,17 @@ rw_lock_sx_lock_low(
 				  thread working on this lock and it is safe to
 				  read and write to the lock_word. */
 
-				ut_ad((lock->lock_word == 0)
-				      || ((lock->lock_word <= -X_LOCK_DECR)
-					  && (lock->lock_word
+#ifdef UNIV_DEBUG
+				int32_t lock_word =
+#endif
+				my_atomic_add32_explicit(&lock->lock_word, -X_LOCK_HALF_DECR,
+							 MY_MEMORY_ORDER_RELAXED);
+
+				ut_ad((lock_word == 0)
+				      || ((lock_word <= -X_LOCK_DECR)
+					  && (lock_word
 					      > -(X_LOCK_DECR
 						  + X_LOCK_HALF_DECR))));
-				lock->lock_word -= X_LOCK_HALF_DECR;
 			}
 		} else {
 			/* Another thread locked before us */
@@ -651,18 +656,12 @@ rw_lock_x_lock_func(
 {
 	ulint		i = 0;
 	sync_array_t*	sync_arr;
-	ulint		spin_count = 0;
-	uint64_t	count_os_wait = 0;
+	lint		spin_count = 0;
+	int64_t		count_os_wait = 0;
 
 	ut_ad(rw_lock_validate(lock));
 	ut_ad(!rw_lock_own(lock, RW_LOCK_S));
 
-	if (rw_lock_x_lock_low(lock, pass, file_name, line)) {
-		/* Locking succeeded */
-		return;
-	}
-	rw_lock_stats.rw_x_spin_wait_count.inc();
-
 lock_loop:
 
 	if (rw_lock_x_lock_low(lock, pass, file_name, line)) {
@@ -682,16 +681,14 @@ lock_loop:
 
 		/* Spin waiting for the lock_word to become free */
 		HMT_low();
-		ulint j = i;
 		while (i < srv_n_spin_wait_rounds
-		       && lock->lock_word <= X_LOCK_HALF_DECR) {
-
+		       && my_atomic_load32_explicit(&lock->lock_word, MY_MEMORY_ORDER_RELAXED) <= X_LOCK_HALF_DECR) {
 			ut_delay(srv_spin_wait_delay);
 			i++;
 		}
 
 		HMT_medium();
-		spin_count += lint(i - j);
+		spin_count += lint(i);
 
 		if (i >= srv_n_spin_wait_rounds) {
 
@@ -710,7 +707,7 @@ lock_loop:
 
 	/* Waiters must be set before checking lock_word, to ensure signal
 	is sent. This could lead to a few unnecessary wake-up signals. */
-	my_atomic_fas32((int32*) &lock->waiters, 1);
+	my_atomic_fas32_explicit(&lock->waiters, 1, MY_MEMORY_ORDER_ACQUIRE);
 
 	if (rw_lock_x_lock_low(lock, pass, file_name, line)) {
 		sync_array_free_cell(sync_arr, cell);
@@ -757,19 +754,13 @@ rw_lock_sx_lock_func(
 {
 	ulint		i = 0;
 	sync_array_t*	sync_arr;
-	ulint		spin_count = 0;
-	uint64_t	count_os_wait = 0;
+	lint		spin_count = 0;
+	int64_t		count_os_wait = 0;
+	lint		spin_wait_count = 0;
 
 	ut_ad(rw_lock_validate(lock));
 	ut_ad(!rw_lock_own(lock, RW_LOCK_S));
 
-	if (rw_lock_sx_lock_low(lock, pass, file_name, line)) {
-		/* Locking succeeded */
-		return;
-	}
-
-	rw_lock_stats.rw_sx_spin_wait_count.inc();
-
 lock_loop:
 
 	if (rw_lock_sx_lock_low(lock, pass, file_name, line)) {
@@ -781,22 +772,23 @@ lock_loop:
 		}
 
 		rw_lock_stats.rw_sx_spin_round_count.add(spin_count);
+		rw_lock_stats.rw_sx_spin_wait_count.add(spin_wait_count);
 
 		/* Locking succeeded */
 		return;
 
 	} else {
 
-		/* Spin waiting for the lock_word to become free */
-		ulint j = i;
-		while (i < srv_n_spin_wait_rounds
-		       && lock->lock_word <= X_LOCK_HALF_DECR) {
+		++spin_wait_count;
 
+		/* Spin waiting for the lock_word to become free */
+		while (i < srv_n_spin_wait_rounds
+		       && my_atomic_load32_explicit(&lock->lock_word, MY_MEMORY_ORDER_RELAXED) <= X_LOCK_HALF_DECR) {
 			ut_delay(srv_spin_wait_delay);
 			i++;
 		}
 
-		spin_count += lint(i - j);
+		spin_count += lint(i);
 
 		if (i >= srv_n_spin_wait_rounds) {
 
@@ -815,7 +807,7 @@ lock_loop:
 
 	/* Waiters must be set before checking lock_word, to ensure signal
 	is sent. This could lead to a few unnecessary wake-up signals. */
-	my_atomic_fas32((int32*) &lock->waiters, 1);
+	my_atomic_fas32_explicit(&lock->waiters, 1, MY_MEMORY_ORDER_ACQUIRE);
 
 	if (rw_lock_sx_lock_low(lock, pass, file_name, line)) {
 
@@ -828,6 +820,7 @@ lock_loop:
 		}
 
 		rw_lock_stats.rw_sx_spin_round_count.add(spin_count);
+		rw_lock_stats.rw_sx_spin_wait_count.add(spin_wait_count);
 
 		/* Locking succeeded */
 		return;
@@ -853,13 +846,15 @@ rw_lock_validate(
 /*=============*/
 	const rw_lock_t*	lock)	/*!< in: rw-lock */
 {
-	lint	lock_word;
+	int32_t	lock_word;
 
 	ut_ad(lock);
 
-	lock_word = lock->lock_word;
+	lock_word = my_atomic_load32_explicit(const_cast<int32_t*>(&lock->lock_word),
+					      MY_MEMORY_ORDER_RELAXED);
 
-	ut_ad(lock->waiters < 2);
+	ut_ad(my_atomic_load32_explicit(const_cast<int32_t*>(&lock->waiters),
+					MY_MEMORY_ORDER_RELAXED) < 2);
 	ut_ad(lock_word > -(2 * X_LOCK_DECR));
 	ut_ad(lock_word <= X_LOCK_DECR);
 
@@ -922,15 +917,17 @@ rw_lock_add_debug_info(
 	rw_lock_debug_mutex_exit();
 
 	if (pass == 0 && lock_type != RW_LOCK_X_WAIT) {
+		int32_t lock_word = my_atomic_load32_explicit(&lock->lock_word,
+							      MY_MEMORY_ORDER_RELAXED);
 
 		/* Recursive x while holding SX
 		(lock_type == RW_LOCK_X && lock_word == -X_LOCK_HALF_DECR)
 		is treated as not-relock (new lock). */
 
 		if ((lock_type == RW_LOCK_X
-		     && lock->lock_word <  -X_LOCK_HALF_DECR)
+		     && lock_word <  -X_LOCK_HALF_DECR)
 		    || (lock_type == RW_LOCK_SX
-		       && (lock->lock_word < 0 || lock->sx_recursive == 1))) {
+		       && (lock_word < 0 || lock->sx_recursive == 1))) {
 
 			sync_check_lock_validate(lock);
 			sync_check_lock_granted(lock);
@@ -987,7 +984,7 @@ rw_lock_remove_debug_info(
 Checks if the thread has locked the rw-lock in the specified mode, with
 the pass value == 0.
 @return TRUE if locked */
-ibool
+bool
 rw_lock_own(
 /*========*/
 	rw_lock_t*	lock,		/*!< in: rw-lock */
@@ -1019,12 +1016,12 @@ rw_lock_own(
 			rw_lock_debug_mutex_exit();
 			/* Found! */
 
-			return(TRUE);
+			return(true);
 		}
 	}
 	rw_lock_debug_mutex_exit();
 
-	return(FALSE);
+	return(false);
 }
 
 /** Checks if the thread has locked the rw-lock in the specified mode, with
@@ -1107,12 +1104,12 @@ rw_lock_list_print_info(
 
 		count++;
 
-		if (lock->lock_word != X_LOCK_DECR) {
+		if (my_atomic_load32_explicit(const_cast<int32_t*>(&lock->lock_word), MY_MEMORY_ORDER_RELAXED) != X_LOCK_DECR) {
 
 			fprintf(file, "RW-LOCK: %p ", (void*) lock);
 
-			if (lock->waiters) {
-				fputs(" Waiters for the lock exist\n", file);
+			if (int32_t waiters= my_atomic_load32_explicit(const_cast<int32_t*>(&lock->waiters), MY_MEMORY_ORDER_RELAXED)) {
+				fprintf(file, " (%d waiters)\n", waiters);
 			} else {
 				putc('\n', file);
 			}
